local addonName, addon = ...
local L = addon.L

HBD = LibStub("HereBeDragons-2.0")

local QuestieDB = QuestieLoader and QuestieLoader:ImportModule("QuestieDB");
local ZoneDB = QuestieLoader and QuestieLoader:ImportModule("ZoneDB")

local function bit(p)
  return 2 ^ (p - 1)  -- 1-based indexing
end

-- Typical call:  if hasbit(x, bit(3)) then ...
local function hasbit(x, p)
  return x % (p + p) >= p       
end

function addon.isQuestieInstalled()
	return QuestieLoader ~= nil
end

local function checkQuestie()
	if addon.waitingForQuestie or not addon.isQuestieInstalled() then return false end
	if QuestieDB == nil or QuestieDB.QueryQuest == nil then
		if addon.debugging then print("LIME: Questie is not yet initialized") end
		addon.waitingForQuestie = true
		C_Timer.After(4, function()
			addon.waitingForQuestie = false
			addon.loadCurrentGuide()
			addon.updateFromQuestLog()
			addon.updateSteps()
		end)
		return false	
	end
	return true
end

function addon.getQuestNameQuestie(id)
	if id == nil or not checkQuestie() then return end
	local quest = QuestieDB:GetQuest(id)
	if quest == nil then return end
	return quest.name
end

function addon.getQuestSortQuestie(id)
	if id == nil or not checkQuestie() then return end
	local quest = QuestieDB:GetQuest(id)
	if quest == nil then return end
    local parentZoneID = ZoneDB:GetParentZoneId(quest.zoneOrSort)
	if quest.zoneOrSort > 0 then
		return addon.zoneNames[ZoneDB:GetUiMapIdByAreaId(parentZoneID or quest.zoneOrSort)]
	elseif quest.zoneOrSort < 0 then
		for key, n in pairs(QuestieDB.sortKeys) do
			if quest.zoneOrSort == n then
				return string.lower(key):gsub("^%l", string.upper)
			end
		end
	end
end

function addon.getQuestZoneQuestie(id)
	if id == nil or not checkQuestie() then return end
	local quest = QuestieDB:GetQuest(id)
	if quest == nil then return end
    local parentZoneID = ZoneDB:GetParentZoneId(quest.zoneOrSort)
	return addon.zoneNames[ZoneDB:GetUiMapIdByAreaId(parentZoneID or quest.zoneOrSort)]
end

function addon.getQuestPrequestsQuestie(id)
	if id == nil or not checkQuestie() then return end
	local quest = QuestieDB:GetQuest(id)
	if quest == nil then return end
	return quest.preQuestSingle or quest.preQuestGroup
end

function addon.getQuestOneOfPrequestsQuestie(id)
	if id == nil or not checkQuestie() then return end
	local quest = QuestieDB:GetQuest(id)
	if quest == nil then return end
	return quest.preQuestSingle ~= nil
end

function addon.getQuestTypeQuestie(id)
	if id == nil or not checkQuestie() then return end
	local quest = QuestieDB:GetQuest(id)
	if quest == nil then return end
	if quest:IsDungeonQuest() then return "Dungeon" end
	if quest:IsRaidQuest() then return "Raid" end
	if quest:GetQuestTagInfo() == 1 then return "Group" end
	local _, _, _, _, _, isElite = GetQuestTagInfo(id)
	if isElite then return "Elite" end
end

function addon.getQuestLevelQuestie(id)
	if id == nil or not checkQuestie() then return end
	local quest = QuestieDB:GetQuest(id)
	if quest == nil then return end
	return quest.questLevel
end

function addon.getQuestMinimumLevelQuestie(id)
	if id == nil or not checkQuestie() then return end
	local quest = QuestieDB:GetQuest(id)
	if quest == nil then return end
	return quest.requiredLevel
end

function addon.getQuestNextQuestie(id)
	if id == nil or not checkQuestie() then return end
	local quest = QuestieDB:GetQuest(id)
	if quest == nil then return end
	return quest.nextQuestInChain
end

function addon.getQuestRacesQuestie(id)
	if id == nil or not checkQuestie() then return end
	local quest = QuestieDB:GetQuest(id)
	if quest == nil then return end
	local bitmask = quest.requiredRaces
	if bitmask == nil or bitmask == 0 then return end
	local races = {}
	for i, race in ipairs({"Human", "Orc", "Dwarf", "NightElf", "Undead", "Troll", "Gnome", "Tauren", "", "BloodElf", "Draenei"}) do
		if race ~= "" and hasbit(bitmask, bit(i)) then 
			table.insert(races, race) 
		end
	end
	return races
end

function addon.getQuestClassesQuestie(id)
	if id == nil or not checkQuestie() then return end
	local quest = QuestieDB:GetQuest(id)
	if quest == nil then return end
	local bitmask = quest.requiredClasses
	if bitmask == nil or bitmask == 0 then return end
	local classes = {}
	for i, class in pairs({"Warrior", "Paladin", "Hunter", "Rogue", "Priest", "", "Shaman", "Mage", "Warlock", "", "Druid"}) do
		if class ~= "" and hasbit(bitmask, bit(i)) then 
			table.insert(classes, class) 
		end
	end
	return classes
end

function addon.getQuestFactionQuestie(id)
	if id == nil or not checkQuestie() then return end
	local quest = QuestieDB:GetQuest(id)
	if quest == nil then return end
	local bitmask = quest.requiredRaces
	if bitmask == nil then return end
	if bitmask == 77 or bitmask == 1101 then return "Alliance" end
	if bitmask == 178 or bitmask == 690 then return "Horde" end
end

function addon.getQuestObjectiveQuestie(id)
	if id == nil or not checkQuestie() then return end
	local quest = QuestieDB:GetQuest(id)
	if quest == nil then return end
	return quest.objectivesText[1]
end

function addon.getQuestReputationQuestie(id, typ)
	if id == nil or not checkQuestie() then return end
	local quest = QuestieDB:GetQuest(id)
	if quest == nil then return end
	local reputation, repMin, repMax
	if quest.requiredMinRep ~= nil then
		reputation = quest.requiredMinRep[1]
		repMin = quest.requiredMinRep[2]
	end
	if quest.requiredMaxRep ~= nil and (reputation == nil or reputation == quest.requiredMaxRep[1]) then
		reputation = quest.requiredMinRep[1]
		repMax = quest.requiredMinRep[2]
	end
	return reputation, repMin, repMax
end
	

function addon.getQuestPositionsQuestie(id, typ, index, filterZone)
	if id == nil or not checkQuestie() then return end
	local ids = {npc = {}, object = {}, item = {}}
	local objectives = {npc = {}, object = {}, item = {}}
	if addon.getSuperCode(typ) == "QUEST" then
		local quest = QuestieDB:GetQuest(id)
		if quest == nil then return end
		local list
		if typ == "ACCEPT" then 
			list = quest.startedBy
		elseif typ == "COMPLETE" then
			list = quest.objectives
		elseif typ == "TURNIN" then
			list = quest.finishedBy
		end
		if list ~= nil then
			--if addon.debugging then print("LIME: getQuestPositionsQuestie " .. typ .. " " .. id .. " " .. addon.show(list)) end
			local c = 0
			for i, type in ipairs({"npc", "object", "item"}) do
				if list[i] ~= nil then 
					for j, v in ipairs(list[i]) do 
						if index == nil or index == 0 or index == c + j then
							local id
							if typ == "COMPLETE" then id = v[1] else id = v end 
							table.insert(ids[type], id)
							if objectives[type][id] == nil then objectives[type][id] = {} end
							table.insert(objectives[type][id], c + j)
						end
					end 
					c = c + #list[i]
				end
			end
			--kill credit objective
			if list[5] ~= nil then
				c = c + 1
				if index == nil or index == 0 or index == c then
					for j, id in ipairs(list[5][1]) do
						table.insert(ids.npc, id)
						if objectives.npc[id] == nil then objectives.npc[id] = {} end
						table.insert(objectives.npc[id], c)
					end
				end
			end
		end
	elseif typ == "COLLECT_ITEM" then
		ids.item = {id}
		objectives.item[id] = {}
	else
		return
	end
	for j = 1, #(ids.item) do
		local item = QuestieDB:GetItem(ids.item[j])
		--if item == nil then error("item " .. ids.item[j] .. " not found for quest " .. questid .. typ) end
		--if addon.debugging then print("LIME: item", ids.item[j] .. " " .. item[6]) end
		if item ~= nil then
			if item.npcDrops ~= nil then
				for i = 1, #item.npcDrops do
					if not addon.contains(ids.npc, item.npcDrops[i]) then table.insert(ids.npc, item.npcDrops[i]) end
					if objectives.npc[item.npcDrops[i]] == nil then objectives.npc[item.npcDrops[i]] = {} end
					for _, c in ipairs(objectives.item[ids.item[j]]) do table.insert(objectives.npc[item.npcDrops[i]], c) end
				end
			end
			if item.objectDrops ~= nil then
				for i = 1, #item.objectDrops do
					if not addon.contains(ids.object, item.objectDrops[i]) then table.insert(ids.object, item.objectDrops[i]) end
					if objectives.object[item.objectDrops[i]] == nil then objectives.object[item.objectDrops[i]] = {} end
					for _, c in ipairs(objectives.item[ids.item[j]]) do table.insert(objectives.object[item.objectDrops[i]], c) end
				end
			end
		end
	end
	local positions = {}
	local filterZoneId
	if filterZone ~= nil then filterZoneId = ZoneDB:GetAreaIdByUiMapId(addon.mapIDs[filterZone]) end
	for j = 1, #(ids.npc) do
		local npc = QuestieDB:GetNPC(ids.npc[j])
		--if npc == nil then error("npc " .. ids.npc[j] .. " not found for quest " .. questid .. typ) end
		--if addon.debugging then print("LIME: npc", npc[1]) end
		if npc ~= nil and npc.spawns ~= nil then
			if filterZone == nil then
				for zone, posList in pairs(npc.spawns) do
					for _, pos in ipairs(posList) do
						table.insert(positions, {x = pos[1], y = pos[2], zone = addon.zoneNames[ZoneDB:GetUiMapIdByAreaId(zone)] or zone, npcId = ids.npc[j], objectives = objectives.npc[ids.npc[j]]})
					end
				end
			elseif npc.spawns[filterZoneId] ~= nil then
				for _, pos in ipairs(npc.spawns[filterZoneId]) do
					table.insert(positions, {x = pos[1], y = pos[2], zone = filterZone, npcId = ids.npc[j], objectives = objectives.npc[ids.npc[j]]})
				end
			end
		end
	end
	for j = 1, #(ids.object) do
		local object = QuestieDB:GetObject(ids.object[j])
		if object == nil then error("object " .. ids.object[j] .. " not found for quest " .. id .. typ) end
		--if addon.debugging then print("LIME: object", object[1]) end
		if object.spawns ~= nil then
			if filterZone == nil then
				for zone, posList in pairs(object.spawns) do
					for _, pos in ipairs(posList) do
						table.insert(positions, {x = pos[1], y = pos[2], zone = addon.zoneNames[ZoneDB:GetUiMapIdByAreaId(zone)] or zone, objectId = ids.object[j], objectives = objectives.object[ids.object[j]]})
					end
				end
			elseif object.spawns[filterZoneId] ~= nil then
				for _, pos in ipairs(object.spawns[filterZoneId]) do
					table.insert(positions, {x = pos[1], y = pos[2], zone = filterZone, objectId = ids.object[j], objectives = objectives.object[ids.object[j]]})
				end
			end
		end
	end
	local i = 1
	while i <= #positions do
		local pos = positions[i]
		if pos.wx == -1 and pos.wy == -1 then
			-- locations inside instances are marked with -1,-1
			table.remove(positions, i)
		else
			pos.mapID = addon.mapIDs[pos.zone]
			pos.wx, pos.wy, pos.instance = HBD:GetWorldCoordinatesFromZone(pos.x / 100, pos.y / 100, pos.mapID)
			--if addon.debugging then print("LIME: found position", pos.wx, pos.wy, pos.instance) end
			if pos.wx == nil then
				if addon.debugging then print("LIME: error transforming (", pos.x, ",", pos.y, pos.zone, addon.mapIDs[pos.zone], ") into world coordinates for quest #", id) end
				table.remove(positions, i)
			else
				i = i + 1
			end
		end
	end
	--if addon.debugging then print("LIME: found ", #positions, "positions") end
	return positions
end

-- returns a type (npc/item/object) and a list of names for quest source / each objective / turn in; e.g. {{type="item", names={"Huge Gnoll Claw", "Hogger"}, ids={item={1931},npc={448}} for id = 176, typ = "COMPLETE"
function addon.getQuestObjectivesQuestie(id, typ)
	if id == nil or not checkQuestie() then return end
	local quest = QuestieDB:GetQuest(id)
	if quest == nil then return end
	if typ == "ACCEPT" then 
		list = quest.startedBy
	elseif typ == "COMPLETE" then
		list = quest.objectives
	elseif typ == "TURNIN" then
		list = quest.finishedBy
	else
		return
	end
	--if addon.debugging then print("LIME: getQuestObjectivesQuestie " .. typ .. " " .. id .. " " .. addon.show(list)) end
	local objectives = {}
	if list[1] ~= nil then
		for j = 1, #list[1] do
			local objList = {}
			local ids = {}
			local npcId
			if type(list[1][j]) == "number" then 
				npcId = list[1][j]
			else 
				npcId = list[1][j][1]
				table.insert(objList, list[1][j][2])
			end
			local npc = QuestieDB:GetNPC(npcId)
			if npc ~= nil and not addon.contains(objList, npc.name) then table.insert(objList, npc.name) end
			if typ == "COMPLETE" then
				table.insert(objectives, {type = "monster", names = objList, ids = {npc = {npc.id}}})
			else
				table.insert(objectives, {type = "npc", names = objList, ids = {npc = {npcId}}})
			end
		end
	end
	if list[2] ~= nil then
		for j = 1, #list[2] do
			local objList = {}
			local objId
			if type(list[2][j]) == "number" then 
				objId = list[2][j]
			else 
				objId = list[2][j][1]
				table.insert(objList, list[2][j][2])
			end
			local obj = QuestieDB:GetObject(objId)
			if obj ~= nil and not addon.contains(objList, obj.name) then table.insert(objList, obj.name) end
			table.insert(objectives, {type = "object", names = objList, ids = {object = {objId}}})
		end
	end
	if list[3] ~= nil then
		for j = 1, #list[3] do
			local objective = {type = "item", names = {}}
			local itemId 
			if type(list[3][j]) == "number" then 
				itemId = list[3][j]
			else
				itemId = list[3][j][1]
				table.insert(objective.names, list[3][j][2])
			end
			objective.ids = {item = {itemId}}
			local item = QuestieDB:GetItem(itemId)
			if item ~= nil then
				if not addon.contains(objective.names, item.name) then table.insert(objective.names, item.name) end
				if item.npcDrops ~= nil then
					objective.ids.npc = {}
					for i = 1, #item.npcDrops do
						table.insert(objective.ids.npc, item.npcDrops[i])
						local npc = QuestieDB:GetNPC(item.npcDrops[i])
						if npc ~= nil then
							if not addon.contains(objective.names, npc.name) then table.insert(objective.names, npc.name) end
						end
					end
				end
				if item.objectDrops ~= nil then
					objective.ids.object = {}
					for i = 1, #item.objectDrops do
						table.insert(objective.ids.object, item.objectDrops[i])
						local obj = QuestieDB:GetObject(item.objectDrops[i])
						if not addon.contains(objective.names, obj.name) then table.insert(objective.names, obj.name) end
					end
				end
			end
			table.insert(objectives, objective)
		end
	end
	--TODO: reputation objective
	--kill credit objective
	if list[5] ~= nil then
		local objList = {}
		local ids = {}
		local npcIds = list[5][1]
		local npcBaseId = list[5][2]
		local npc = QuestieDB:GetNPC(npcBaseId)
		if npc ~= nil then table.insert(objList, npc.name) end
		if typ == "COMPLETE" then
			table.insert(objectives, {type = "monster", names = objList, ids = {npc = npcIds}})
		else
			table.insert(objectives, {type = "npc", names = objList, ids = {npc = npcIds}})
		end
	end
	return objectives
end

function addon.getNPCPosition(id)
	if id == nil or not checkQuestie() then return end
	local npc = QuestieDB:GetNPC(id)
	if npc ~= nil and npc.spawns ~= nil then
		for zone, posList in pairs(npc.spawns) do
			for _, pos in ipairs(posList) do
				local p = {x = pos[1], y = pos[2], mapID = ZoneDB:GetUiMapIdByAreaId(zone), zone = addon.zoneNames[ZoneDB:GetUiMapIdByAreaId(zone)] or zone}
				p.wx, p.wy, p.instance = HBD:GetWorldCoordinatesFromZone(p.x / 100, p.y / 100, p.mapID)
				if p.wx == nil then
					if addon.debugging then print("LIME: error transforming (", p.x, ",", p.y, p.zone, p.mapID, ") into world coordinates for npc #", id) end
				else
					return p
				end
			end
		end
	end
end
